//****************************************************************************
//
//! \copyright 2020 by WinSystems Inc.
//!
//!   Permission is hereby granted to the purchaser of WinSystems  cards
//!   and CPU products incorporating a GPIO device, to distribute any binary
//!   file or files compiled using this source code directly or in any work
//!   derived by the user from this file. In no case may the source code,
//!   original or derived from this file, be distributed to any third party
//!   except by explicit permission of WinSystems. This file is distributed
//!   on an "As-is" basis and no warranty as to performance or fitness of pur-
//!   poses is expressed or implied. In no case shall WinSystems be liable for
//!   any direct or indirect loss or damage, real or consequential resulting
//!   from the usage of this source code. It is the user's sole responsibility
//!   to determine fitness for any considered purpose.
//
///****************************************************************************
//
//    Filename   : i440GetSample.c
//
//    Project    : i440 Example Applications
//
//! \author     pbd
//
///****************************************************************************
//
//! \date       07/28/2020
//
//
//! \version    1.0
//
//
///****************************************************************************

#define _DEFAULT_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <errno.h>

#define I440_PERF    1

#if I440_PERF == 1
#include <sys/time.h>
#endif

#include <x86_64_types.h>
#include <libi440.h>

#define APP_NAME        "i440GetSamples"
#define MAJOR_VER       1
#define MINOR_VER       0
#define TIME_BASE       599


//! \cond

void print_usage(void)
{
    printf("Usage error:\n");
    printf("  i440GetSamples <Channel> <Count> <Timebase>\n");
    printf("  Channel must be between 0 and 7\n");
    printf("  Count can be any number less than 65536\n");
    printf("  Timebase can be any number greater than 58\n");
    printf("  Timebase is optional, default value is 59\n");
}

//! \endcond

//****************************************************************************
//
//! \fn          main
//
//! \brief     Reads a specified count of ADC samples from the specified ADC
//!            channel, using the specified time base
//
//! @param[in]   <CHAN>      ADC channel to read from
//! @param[in]   <CNT>       Count of samples to read from specified channel
//! @param[in]   <TIME>      Timebase to use while reading samples. See the
//!                          enumeration TimeBase in file libi440.h
//
//! \return      int     Integer value indicating success or failure. If 0,
//!                      operation succeeded in a normal manner, otherwise
//!                      an error occurred.
//
//! \note   
//
//****************************************************************************
int main( int argc, char** argv )
{
    uint32_t   TotalBytes = 0;
    uint32_t   Time = TIME_1MHz;
    
    uint32_t   Chan = 0;
    uint32_t   SampleCount = 10;    
    
    int        Status = SUCCESS;

    printf( "I440 Application : %s\n", APP_NAME );
    printf( "Version %d.%d\n\n", MAJOR_VER, MINOR_VER);

    if ( ( argc != 3 ) && ( argc != 4) )
        {
           print_usage();
           Status = INVALID_PARAMETER;
        }
    else
        {
           //
           // load channel
           //
           
           Chan = atoi( argv[ 1 ] );

           //
           // load count, add one for dummy conversion at start
           //
           
           SampleCount = atoi( argv[ 2 ] ); 

           //
           // load time base if included
           //
           
           if ( argc == 4 )
           {
              Time = atoi( argv[ 3 ] );
           }

           printf( "\n%d samples on channel %d at timebase %d\n", SampleCount, Chan, Time );

           //
           // total memory size required
           //
           
           TotalBytes = ( SampleCount * 3 );
           
           if ( ( Status = InitializeSession() ) == SUCCESS )
               {
                  printf("FT232 Sync FIFO Mode opened.\n");

                  //
                  // configure channel
                  //
                  
                  SetChannelMode( Chan, ADC_BIPOLAR, ADC_TOP_10_24V );
                  
                  //
                  // set time base
                  //
                  
                  Status = SetTimeBase( Time );
                  
                  if ( Status )
                      {
                         printf("Error setting time base!\n\n");
                      }
                  else
                      {
                         //
                         // allocate memory for samples
                         //
                            
                         puint8_t   pBuffer = ( puint8_t ) malloc( TotalBytes );

                         if ( pBuffer == NULL )
                             {
                                printf( "Insufficient memory available!\n\n" );
                                Status = ENOMEM;
                             }
                         else
                             {
#if I440_PERF == 1
                                struct timeval     StartTime,
                                                   EndTime,
                                                   ElapsedTime;
                                bool               TimeReadOk = false;
#endif                                
                                
                                printf( "%d bytes of memory space allocated for sample storage.\n\n", TotalBytes );
                                   
#if I440_PERF == 1
                                if ( gettimeofday( &StartTime, NULL ) == 0 )
                                {
                                   TimeReadOk = true;   
                                }
#endif
                                //
                                // get samples
                                //
                                   
                                printf( "\nGet %d samples for channel %d\n", SampleCount, Chan );
                                Status = GetChannelDataCount( pBuffer, Chan, &SampleCount );

                                if ( Status )
                                    {
                                       printf( "\nError %d executing conversions for channel %d!\n", Status, Chan );
                                       FlushRxQueue();
                                    }
                                else
                                    {
                                       float       ConvVoltage;
                                       uint32_t    RawData;
#if I440_PERF == 1
                                       if ( TimeReadOk )
                                       {
                                          if ( gettimeofday( &EndTime, NULL ) == 0 )
                                          {
                                             timersub( &EndTime, &StartTime, &ElapsedTime );
                                             printf ("Read %d samples in %ld.%06ld seconds\n", SampleCount, ElapsedTime.tv_sec, ElapsedTime.tv_usec);
                                          }
                                       }
#endif
                                       //
                                       // display samples as voltages
                                       //
                                          
                                       for ( uint32_t i = 0; i < TotalBytes; i += 3 )
                                       {
                                          Chan = ( *( pBuffer + i + 2 ) >> 3 ) & 0x7;
                                          RawData = ( *( pBuffer + i) << 8) | *( pBuffer + i + 1 );

                                          ConvertToVolts( &ConvVoltage, Chan, RawData );
                                          printf("Sample %5d - Channel %d - Voltage = %.3f\n", i / 3 + 1, Chan, ConvVoltage);
                                       }
                                    }
                                //
                                // free the buffer
                                //
                            
                                free( pBuffer );
                             }
                      }
                  
                  //             
                  // Close device
                  //
                  
                  Status = CloseSession();
                  
                  if ( Status )
                      {
                         printf( "Error closing device.\n" );
                      }
                  else
                      {
                         printf("FT232 device closed.\n");   
                      }
               }
           else
               {
                  printf("Error initializing device.\n");
               }
        }

   return ( Status );
}
