//****************************************************************************
//
//! \copyright 2020 by WinSystems Inc.
//!
//!   Permission is hereby granted to the purchaser of WinSystems  cards
//!   and CPU products incorporating a GPIO device, to distribute any binary
//!   file or files compiled using this source code directly or in any work
//!   derived by the user from this file. In no case may the source code,
//!   original or derived from this file, be distributed to any third party
//!   except by explicit permission of WinSystems. This file is distributed
//!   on an "As-is" basis and no warranty as to performance or fitness of pur-
//!   poses is expressed or implied. In no case shall WinSystems be liable for
//!   any direct or indirect loss or damage, real or consequential resulting
//!   from the usage of this source code. It is the user's sole responsibility
//!   to determine fitness for any considered purpose.
//
///****************************************************************************
//
//    Filename   : i440GetSequence.c
//
//    Project    : i440 Example Applications
//
//! \author     pbd
//
///****************************************************************************
//
//! \date       07/28/2020
//
//
//! \version    1.0
//
//
///****************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <errno.h>
#include <x86_64_types.h>
#include <libi440.h>

#define APP_NAME    "i440GetSequence"
#define MAJOR_VER       1
#define MINOR_VER       0
#define TIME_BASE       599

//! \cond

void print_usage(void)
{
    printf("Usage error:\n");
    printf("  i440GetSequence <Timebase>\n");
    printf("  Timebase can be any number greater than 58\n");
    printf("  Timebase is optional, default value is 59\n");
}

//! \endcond

//****************************************************************************
//
//! \fn        main
//
//! \brief     Reads a sequence of channels. The sequence is hard coded in the
//!            array chSeq. An optional time base may be provided
//!            channel, using the specified time base
//
//! @param[in]   <TIME>      Timebase to use while reading samples. See the
//!                          enumeration TimeBase in file libi440.h THIS
//!                          PARAMETER IS OPTIONAL, AND IF NOT PROVIDED, WILL
//!                          DEFAULT TO THE ENUMERATION VALUE FOR 1 MHz
//
//! \return      int     Integer value indicating success or failure. If 0,
//!                      operation succeeded in a normal manner, otherwise
//!                      an error occurred.
//
//! \note   The optional parameter Timebase must be greater than 59, which will
//!         set the sampling rate to 1 MHz or slower. See TimeBase in file
//!         libi440.h
//
//****************************************************************************
int main( int argc, char** argv )
{
    uint8_t    cmd[ 16 ] = { 0 };
    
    uint32_t chSeq[] = { 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 0, 0xff };
    uint32_t Size = sizeof( chSeq ) / sizeof( uint32_t ) - 1;
    uint32_t Time = 59;
        
    uint32_t   TotalBytes = 0;
    uint32_t   TotalConvs = 0;
    
    uint32_t ch = 0;
    uint32_t dup = 0;
    uint32_t rng = 0;
    uint32_t data;
    
    puint8_t pBuffer;
    
    int Status = SUCCESS;

    printf("I440 Application : %s\n", APP_NAME );
    printf("Version %d.%d\n\n", MAJOR_VER, MINOR_VER);

    if ( argc > 2 )
        {
           print_usage();
           Status = INVALID_PARAMETER;
        }
    else
        {
           //
           // total memory size required
           //
           
           TotalBytes = Size * 3;
           
           if ( ( Status = InitializeSession() ) == SUCCESS )
               {
                  printf("FT232 Sync FIFO Mode opened.\n");
                  
                  //
                  // configure all channels
                  //
                  
                  for (int i = 0; i < 8; i++)
                  {
                     Status = SetChannelMode(i, ADC_BIPOLAR, ADC_TOP_10_24V);
        
                     if ( Status )
                     {
                        printf("\nError configuring channel %d!\n", i);
                        break;
                     }
                  }
                  
                  if ( Status == SUCCESS )
                  {
                     //
                     // set time base
                     //
                     
                     Status = SetTimeBase( Time );

                     if ( Status == SUCCESS )
                        {
                           //
                           // allocate memory for samples
                           //
                           
                           pBuffer = ( puint8_t) malloc( TotalBytes );

                           if ( pBuffer == NULL )
                               {
                                  printf("\nInsufficient memory available!\n");
                                  Status = ENOMEM;
                               }
                           else
                               {
                                  printf("\n%d bytes of memory space allocated for sample storage.\n\n", TotalBytes );                      
                                 
                                  //
                                  // get samples
                                  //
                                  
                                  Status = GetChannelSequence( pBuffer, chSeq );

                                  if ( Status )
                                      {
                                         printf("\nError %d executing sequence!\n", Status );
                                      }
                                  else
                                      {
                                         float ConvVoltage;                                         
                                         
                                         //
                                         // display samples as voltages
                                         //
                                         
                                         for ( uint32_t i = 0; i < ( Size * 3 ); i += 3 )
                                         {
                                            ch = (*( pBuffer + i + 2 ) >> 3 ) & 0x7;
                                            data = (*( pBuffer + i ) << 8) | *( pBuffer + i + 1 );

                                            ConvertToVolts( &ConvVoltage, ch, data);
                                            printf("Sample %5d - Channel %d - Voltage = %.3f\n", ( i / 3 + 1 ), ch, ConvVoltage);
                                         }
                                      }
                                  
                                  //
                                  // free the buffer
                                  //
                            
                                  free( pBuffer );
                               }
                        }
                    else
                        {
                           printf("\nError setting time base!\n");
                        }
                  }
              
                  //             
                  // Close device
                  //
                  
                  Status = CloseSession();
                  
                  if ( Status )
                      {
                         printf( "Error closing device.\n" );
                      }
                  else
                      {
                         printf("FT232 device closed.\n");   
                      }
               }
           else
               {
                  printf("Error initializing device.\n");
               }
        }

   return ( Status );
}

