//****************************************************************************
//	
//	Copyright 2020 by WinSystems Inc.
//
//****************************************************************************
//
//	Name	 : libi440.cpp
//
//	Project	 : PX1-I440-ADC Linux Device Driver
//
//****************************************************************************
//
//	  Date		Revision	                Description
//	--------	--------	---------------------------------------------
//	07/24/20	  1.0		Original Release	
//
//****************************************************************************

#include "libi440.h"
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <string>
#include "ftd2xx.h"
#include "WinTypes.h"

#define I440_DEBUG          0
#define I440_ERR            1
#define I440_VID	    0x328c
#define I440_PID	    0x0440
#define VALID_BIT           (1 << 7)
#define TIMEOUT             5
#define USB_LATCY_TMR       20
#define USB_IN_XFER_SIZE    65536
#define ALL_OUTPUTS         0xFF
#define MAX_SEQ_SIZE        16
#define MAX_CMD_SIZE        512
#define debug_print(...)    if (I440_DEBUG) fprintf(stderr, __VA_ARGS__);
#define err_print(...)      if (I440_ERR) fprintf(stderr, __VA_ARGS__);

#define TIMED_WAIT          Status = pthread_cond_timedwait(&eh.eCondVar, &eh.eMutex, &tout);\
                            debug_print("Event Status = %08x\n", Status);\
                            if (Status == ETIMEDOUT)\
                            {\
                                debug_print("FT232 time-out waiting for RX event!\n");\
                                MUTEX_UNLOCK\
                                return Status = TIMEOUT_ERROR;\
                            }\
                            else if (Status != WAIT_OBJECT_0)\
                            {\
                                debug_print("Error waiting for RX event!\n");\
                                MUTEX_UNLOCK\
                                return Status = DRIVER_ERROR;\
                            }

#define MUTEX_LOCK          Status = pthread_mutex_lock(&eh.eMutex);\
                            if (Status != FT_OK)\
                            {\
                                debug_print("Mutex lock failed!\n");\
                                return Status = DRIVER_ERROR;\
                            }

#define MUTEX_UNLOCK        Status = pthread_mutex_unlock(&eh.eMutex);\
                            if (Status != FT_OK)\
                            {\
                                debug_print("Mutex unlock failed!\n");\
                                return Status = DRIVER_ERROR;\
                            }

#define GET_TIME            Status = clock_gettime(CLOCK_REALTIME, &tout);\
                            if (Status != FT_OK)\
                            {\
                                debug_print("Get time failed!\n");\
                                return Status = DRIVER_ERROR;\
                            }\
                            else\
                                tout.tv_sec += TIMEOUT;

typedef enum {
    CMD_LOAD_SEQ = 0x40,
    CMD_AUTO_CNV = 0x41,
    CMD_CONFIG = 0x42,
    CMD_SEQ_CNT = 0x4C,
    CMD_TIME_BASE = 0x4D,
    CMD_STOP_AUTO = 0x4E,
    CMD_START_AUTO = 0x4F,
    CMD_FPGA_REV = 0x50,
    CMD_SPI_OPCODE = 0x51
} FpgaCmd;

static FT_HANDLE i440Handle = NULL;
static UINT8 chanSoftSpan[8] = { 0 };
static UINT16 chanAdjust[8] = { 0 };
static float chanBitVal[8] = { 0.0 };
static float chanOffset[8] = { 0.0 };
static EVENT_HANDLE eh;

extern "C"
{
    int InitializeSession(void)
    {
        debug_print("Initialize Session started\n");
        BYTE Mask = ALL_OUTPUTS; // data bus all outputs
        UCHAR LatencyTimer = USB_LATCY_TMR;
        DWORD InTransferSize = USB_IN_XFER_SIZE;
        unsigned int throwawayRead;
        int Status = SUCCESS;

        Status = FT_SetVIDPID(I440_VID, I440_PID);

        if (Status != FT_OK)
        {
            debug_print("Failed to set vid and pid %d\n", Status);
            return Status;
        }
        else
            debug_print("Vid and Pid are set ... \n");

        // open the device
        Status = FT_Open(0, &i440Handle);

        if (Status != FT_OK)
        {
            debug_print("Device failed to open %d\n", Status);
            return Status;
        }
        else
            debug_print("Device opened ... \n");

        // set latency timer
        Status = FT_SetLatencyTimer(i440Handle, LatencyTimer);

        if (Status != FT_OK)
            goto Exit;

        // what is latency timer?
        Status = FT_GetLatencyTimer(i440Handle, &LatencyTimer);

        if (Status != FT_OK)
            goto Exit;
        else
            debug_print("Latency timer = %d ms\n", LatencyTimer);

        // enlarge USB FIFO size
        Status = FT_SetUSBParameters(i440Handle, InTransferSize, 0);

        if (Status != FT_OK)
            goto Exit;
        else
            debug_print("USB FIFO size set to %d\n", InTransferSize);

        // reset the device
        Status = FT_ResetDevice(i440Handle);

        if (Status != FT_OK)
            goto Exit;
        else
            debug_print("Device reset ... ");

        // reset
        Status = FT_SetBitMode(i440Handle, Mask, FT_BITMODE_RESET);

        // set device mode to Sync FIFO
        Status = FT_SetBitMode(i440Handle, Mask, FT_BITMODE_SYNC_FIFO);

        if (Status != FT_OK)
            goto Exit;
        else
            debug_print("Sync FIFO mode set!\n");

        Status = GetFpgaRevision(&throwawayRead);

        if (Status != FT_OK)
            goto Exit;
        else
            debug_print("Throwaway transaction done!\n");

        return Status;

    Exit:
        FT_Close(i440Handle);
        return Status;
    }

    int CloseSession(void)
    {
        int Status = SUCCESS;

        Status = FT_Close(i440Handle);

        if (Status != FT_OK)
            debug_print("... device closed!\n");

        return Status;
    }

    int GetFpgaRevision(unsigned int *fpgaRevision)
    {
        UINT8 readBuf[1] = { 0 };
        UINT8 writeBuf[1] = { 0 };
        DWORD TxBytes = 0;
        DWORD RxBytes = 0;
        DWORD BytesReceived = 0;
        DWORD EventDWord = 0;
        DWORD EventMask = FT_EVENT_RXCHAR;
        timespec tout;
        int Status = SUCCESS;

        // check parameters
        if (fpgaRevision == NULL)
        {
            Status = INVALID_PARAMETER;
        }
        else
        {
            // flush FIFO
            Status = FlushRxQueue();

            if (Status != FT_OK)
            {
                debug_print("FT232 FIFO flush failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_mutex_init(&eh.eMutex, NULL);

            if (Status != FT_OK)
            {
                debug_print("Mutex initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_cond_init(&eh.eCondVar, NULL);

            if (Status != FT_OK)
            {
                debug_print("Condition variable initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = FT_SetEventNotification(i440Handle, EventMask, (PVOID)&eh);

            if (Status != FT_OK)
            {
                debug_print("FT232 set event notification failed!\n");
                return Status = DRIVER_ERROR;
            }

            // build control word
            writeBuf[0] = CMD_FPGA_REV;

            Status = FT_Write(i440Handle, writeBuf, sizeof(writeBuf), &BytesReceived);

            if ((Status != FT_OK) || (BytesReceived != (UINT16) sizeof(writeBuf)))
            {
                debug_print("FT232 FIFO write of %d bytes failed!\n", (UINT16) sizeof(writeBuf));
                Status = DRIVER_ERROR;
            }
            else
            {
                debug_print("Write of %d bytes successful!\n", BytesReceived);

                // wait for bytes to be received
                GET_TIME
                MUTEX_LOCK
                TIMED_WAIT
                MUTEX_UNLOCK

                Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                if (Status != FT_OK)
                {
                    debug_print("FT232 get status failed!\n");
                    return Status = DRIVER_ERROR;
                }
                else if (RxBytes)
                {
                    debug_print("FT232 status returned: RxBytes = %d bytes\n", RxBytes);

                    Status = FT_Read(i440Handle, readBuf, RxBytes, &BytesReceived);

                    if ((Status != FT_OK) || (BytesReceived != RxBytes))
                    {
                        debug_print("FT232 FIFO read of %d bytes failed!\n", RxBytes);
                        return Status = DRIVER_ERROR;
                    }
                    else
                    {
                        debug_print("Read of %d bytes successful!\n", BytesReceived);
                    }
                }

                *fpgaRevision = readBuf[0];
                debug_print("FPGA revision: %02x\n", readBuf[0]);
            }
        }

        //CloseHandle(hEvent);

        return Status;
    }

    int GetSoftwareVersion(unsigned int *pDriverVersion, unsigned int *pLibVersion)
    {
        int Status = 1;

        return Status;
    }

    int SetChannelMode(unsigned int channel, int duplex, int range)
    {
        int Status = SUCCESS;

        // check parameters
        if ((channel >= ADC_NO_CHANNEL) ||
                ((duplex != ADC_UNIPOLAR) && (duplex != ADC_BIPOLAR)) ||
                ((range != ADC_TOP_5V) && (range != ADC_TOP_5_12V) && (range != ADC_TOP_10V) && (range != ADC_TOP_10_24V)))
        {
            return Status = INVALID_PARAMETER;
        }

        chanSoftSpan[channel] = (BYTE)duplex | (BYTE)range;

        debug_print("Channel %d configured:\n", channel);
        debug_print("  Duplex = 0x%x\n", duplex);
        debug_print("  Range  = 0x%x\n", range);
        debug_print("  Span   = 0x%x\n", chanSoftSpan[channel]);

        // calculate bit values, offset, and adjustment values
        if ((range == ADC_TOP_5V) && (duplex == ADC_UNIPOLAR))
        {
            chanBitVal[channel] = (float) 5.00 / (float) 65535.0;
            chanAdjust[channel] = 0;
            chanOffset[channel] = 0.0;
        }
        else if ((range == ADC_TOP_5V) && (duplex == ADC_BIPOLAR))
        {
            chanBitVal[channel] = (float) 10.00 / (float) 65535.0;
            chanAdjust[channel] = 0x8000;
            chanOffset[channel] = -5.0;
        }
        else if ((range == ADC_TOP_5_12V) && (duplex == ADC_UNIPOLAR))
        {
            chanBitVal[channel] = (float) 5.12 / (float) 65535.0;
            chanAdjust[channel] = 0;
            chanOffset[channel] = 0.0;
        }
        else if ((range == ADC_TOP_5_12V) && (duplex == ADC_BIPOLAR))
        {
            chanBitVal[channel] = (float) 10.24 / (float) 65535.0;
            chanAdjust[channel] = 0x8000;
            chanOffset[channel] = (float)-5.12;
        }

        if ((range == ADC_TOP_10V) && (duplex == ADC_UNIPOLAR))
        {
            chanBitVal[channel] = (float) 10.00 / (float) 65535.0;
            chanAdjust[channel] = 0;
            chanOffset[channel] = 0.0;
        }
        else if ((range == ADC_TOP_10V) && (duplex == ADC_BIPOLAR))
        {
            chanBitVal[channel] = (float) 20.00 / (float) 65535.0;
            chanAdjust[channel] = 0x8000;
            chanOffset[channel] = -10.0;
        }

        if ((range == ADC_TOP_10_24V) && (duplex == ADC_UNIPOLAR))
        {
            chanBitVal[channel] = (float) 10.24 / (float) 65535.0;
            chanAdjust[channel] = 0;
            chanOffset[channel] = 0.0;
        }
        else if ((range == ADC_TOP_10_24V) && (duplex == ADC_BIPOLAR))
        {
            chanBitVal[channel] = (float) 20.48 / (float) 65535.0;
            chanAdjust[channel] = 0x8000;
            chanOffset[channel] = (float)-10.24;
        }

        return Status;
    }

    int SetTimeBase(unsigned int time)
    {
        UINT8 writeBuf[3] = { 0x00 };
        DWORD BytesReceived = 0;
        int Status = SUCCESS;

        // check parameters
        if ((time < TIME_1MHz) || (time > TIME_LIMIT))
        {
            Status = INVALID_PARAMETER;
        }
        else
        {
            // build command fifo
            writeBuf[0] = CMD_TIME_BASE;
            writeBuf[1] = (UINT8)(time & 0xff);
            writeBuf[2] = (UINT8)((time >> 8) & 0xff);

            Status = FT_Write(i440Handle, writeBuf, sizeof(writeBuf), &BytesReceived);

            if ((Status != FT_OK) || (BytesReceived != sizeof(writeBuf)))
            {
                debug_print("FT232 FIFO write of %d bytes failed!\n", (int) sizeof(writeBuf));
                Status = DRIVER_ERROR;
            }
        }

        return Status;
    }

    int GetChannelData(unsigned int *pValue, unsigned int channel)
    {
        UINT8 readBuf[3] = { 0 };
        UINT8 writeBuf[1] = { 0 };
        DWORD TxBytes = 0;
        DWORD RxBytes = 0;
        DWORD BytesReceived = 0;
        DWORD EventDWord = 0;
        DWORD EventMask = FT_EVENT_RXCHAR;
        timespec tout;
        int Status = SUCCESS;

        // check parameters
        if ((pValue == NULL) || (channel >= ADC_NO_CHANNEL))
        {
            Status = INVALID_PARAMETER;
        }
        else
        {
            // flush FIFO
            Status = FlushRxQueue();

            if (Status != FT_OK)
            {
                debug_print("FT232 FIFO flush failed!  %d\n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_mutex_init(&eh.eMutex, NULL);

            if (Status != FT_OK)
            {
                debug_print("Mutex initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_cond_init(&eh.eCondVar, NULL);

            if (Status != FT_OK)
            {
                debug_print("Condition variable initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = FT_SetEventNotification(i440Handle, EventMask, &eh);

            if (Status != FT_OK)
            {
                debug_print("FT232 set event notification failed!\n");
                return Status = DRIVER_ERROR;
            }

            // build control word
            writeBuf[0] = (UINT8)(VALID_BIT | (channel << 3) | chanSoftSpan[channel]);

            Status = FT_Write(i440Handle, writeBuf, sizeof(writeBuf), &BytesReceived);

            if ((Status != FT_OK) || (BytesReceived != (UINT16) sizeof(writeBuf)))
            {
                debug_print("FT232 FIFO write of %d bytes failed!\n", (UINT16) sizeof(writeBuf));
                Status = DRIVER_ERROR;
            }
            else
            {
                debug_print("Write of %d bytes successful!\n", BytesReceived);

                // wait for bytes to be received
                GET_TIME
                MUTEX_LOCK
                TIMED_WAIT
                MUTEX_UNLOCK

                Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                debug_print("RxBytes    = %d\n", RxBytes);
                debug_print("TxBytes    = %d\n", TxBytes);
                debug_print("EventDWord = %08x\n", EventDWord);

                if (Status != FT_OK)
                {
                    debug_print("FT232 get status failed!\n");
                    return Status = DRIVER_ERROR;
                }
                else if (RxBytes)
                {
                    debug_print("FT232 status returned: RxBytes = %d bytes\n", RxBytes);

                    Status = FT_Read(i440Handle, readBuf, RxBytes, &BytesReceived);

                    if ((Status != FT_OK) || (BytesReceived != RxBytes))
                    {
                        debug_print("FT232 FIFO read of %d bytes failed!\n", RxBytes);
                        return Status = DRIVER_ERROR;
                    }
                    else
                    {
                        debug_print("Read of %d bytes successful!\n", BytesReceived);
                    }
                }

                *pValue = (readBuf[0] << 16) | (readBuf[1] << 8) | readBuf[2];

                debug_print("FT232 conversion data:\n");
                debug_print("  D15 - D0 = 0x%04x\n", (readBuf[0] << 8) | readBuf[1]);
                debug_print("  Channel  = %d\n", (readBuf[2] >> 3) & 0x7);
                debug_print("  SoftSpan = %d\n", readBuf[2] & 0x7);
            }
        }

        return Status;
    }

    int GetChannelDataCount(unsigned char *pValBuf, unsigned int channel, unsigned int *pCount)
    {
        UINT8 writeBuf[4] = { 0 };
        DWORD byteCtr = 0;
        DWORD TxBytes = 0;
        DWORD RxBytes = 0;
        DWORD BytesReceived = 0;
        DWORD EventDWord = 0;
        DWORD EventMask = FT_EVENT_RXCHAR;
        timespec tout;
        BOOL noCnt = FALSE;
        int Status = SUCCESS;

        // check parameters
        if ((pValBuf == NULL) || (channel >= ADC_NO_CHANNEL) || (pCount == NULL) || (*pCount == 0))
        {
            Status = INVALID_PARAMETER;
        }
        else
        {
            // flush FIFO
            Status = FlushRxQueue();

            if (Status != FT_OK)
            {
                debug_print("FT232 FIFO flush failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            // is cnt = 0?
            //if (*pCount == 0)
            //    noCnt = TRUE;

            Status = pthread_mutex_init(&eh.eMutex, NULL);

            if (Status != FT_OK)
            {
                debug_print("Mutex initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_cond_init(&eh.eCondVar, NULL);

            if (Status != FT_OK)
            {
                debug_print("Condition variable initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = FT_SetEventNotification(i440Handle, EventMask, &eh);

            if (Status != FT_OK)
            {
                err_print("FT232 set event notification failed!\n");
                return Status = DRIVER_ERROR;
            }

            // build control words and start conversions
            writeBuf[0] = (UINT8)(VALID_BIT | (channel << 3) | chanSoftSpan[channel]);
            writeBuf[1] = CMD_SEQ_CNT;
            writeBuf[2] = (UINT8)(*pCount & 0xff);
            writeBuf[3] = (UINT8)((*pCount >> 8) & 0xff);

            debug_print("Write buffer contents:\n");
            for (unsigned int j = 0; j < 4; j++)
                debug_print(" %02x -", writeBuf[j]);
            debug_print("\n");

            Status = FT_Write(i440Handle, writeBuf, sizeof(writeBuf), &BytesReceived);

            if ((Status != FT_OK) || (BytesReceived != (DWORD)sizeof(writeBuf)))
            {
                debug_print("FT232 FIFO write of %d bytes failed!\n", (UINT16)sizeof(writeBuf));
                Status = DRIVER_ERROR;
            }
            else
            {
                debug_print("Write of %d bytes successful!\n", BytesReceived);

                do {
                    // wait for bytes to be received
                    GET_TIME
                    MUTEX_LOCK

                    while (RxBytes == 0)
                    {
                        debug_print("Starting receive\n");

                        Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                        debug_print("Ending receive\n");

                        if (Status != FT_OK)
                        {
                            err_print("FT232 get status failed!\n");
                            return Status = DRIVER_ERROR;
                        }

                        if (RxBytes > 0)
                            break;

                        TIMED_WAIT
                    }

                    MUTEX_UNLOCK

                    Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                    debug_print("RxBytes    = %d\n", RxBytes);
                    debug_print("TxBytes    = %d\n", TxBytes);
                    debug_print("EventDWord = %08x\n", EventDWord);

                    if (Status != FT_OK)
                    {
                        debug_print("FT232 get status failed!\n");
                        return Status = DRIVER_ERROR;
                    }
                    else if (RxBytes)
                    {
                        Status = FT_Read(i440Handle, pValBuf + byteCtr, RxBytes, &BytesReceived);

                        if ((Status != FT_OK) || (BytesReceived != RxBytes))
                        {
                            debug_print("FT232 FIFO read of %d bytes failed!\n", RxBytes);
                            debug_print("  Status = %d\n", Status);
                            return Status = DRIVER_ERROR;
                        }
                        else
                            debug_print("Read of %d bytes successful!\n", BytesReceived);

                        byteCtr += (UINT16)BytesReceived;
                    }
                } while ((byteCtr < *pCount * 3) || noCnt);

    #if 0
                if (Status == SUCCESS)
                {
                    for (unsigned int i = 0; i < *pCount; i++)
                    {
                        debug_print("FT232 conversion data:\n");
                        debug_print("  Sample %d\n", i + 1);
                        debug_print("  D15 - D0 = 0x%04x\n", (*(pValBuf + (i * 3)) << 8) | *(pValBuf + (i * 3) + 1));
                        debug_print("  Channel  = %d\n", (*(pValBuf + (i * 3) + 2) >> 3) & 0x7);
                        debug_print("  SoftSpan = %d\n", *(pValBuf + (i * 3) + 2) & 0x7);
                    }
                }
    #endif
            }
        }

        return Status;
    }

    int GetAllChannelData(unsigned char *pValBuf)
    {
        UINT8 writeBuf[9] = { 0 };
        DWORD byteCtr = 0;
        DWORD TxBytes = 0;
        DWORD RxBytes = 0;
        DWORD BytesReceived = 0;
        DWORD EventDWord = 0;
        DWORD EventMask = FT_EVENT_RXCHAR;
        timespec tout;
        int Status = SUCCESS;

        // check parameters
        if (pValBuf == NULL)
        {
            Status = INVALID_PARAMETER;
        }
        else
        {
            // flush FIFO
            Status = FlushRxQueue();

            if (Status != FT_OK)
            {
                debug_print("FT232 FIFO flush failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_mutex_init(&eh.eMutex, NULL);

            if (Status != FT_OK)
            {
                debug_print("Mutex initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_cond_init(&eh.eCondVar, NULL);

            if (Status != FT_OK)
            {
                debug_print("Condition variable initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = FT_SetEventNotification(i440Handle, EventMask, &eh);

            if (Status != FT_OK)
            {
                debug_print("FT232 set event notification failed!\n");
                return Status = DRIVER_ERROR;
            }

            // write FIFO data
            for (int i = 0; i < 8; i++)
                writeBuf[i] = (UINT8)(VALID_BIT | (i << 3) | chanSoftSpan[i]);

            Status = FT_Write(i440Handle, writeBuf, sizeof(writeBuf), &BytesReceived);

            if ((Status != FT_OK) || (BytesReceived != (UINT16) sizeof(writeBuf)))
            {
                debug_print("FT232 FIFO write of %d bytes failed!\n", (UINT16) sizeof(writeBuf));
                Status = DRIVER_ERROR;
            }
            else
            {
                debug_print("Write of %d bytes successful!\n", BytesReceived);
                do {
                    Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                    if (Status != FT_OK)
                    {
                        err_print("FT232 get status failed!\n");
                        return Status = DRIVER_ERROR;
                    }

                    while (RxBytes == 0)
                    {
                        Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                        if (Status != FT_OK)
                        {
                            err_print("FT232 get status failed!\n");
                            return Status = DRIVER_ERROR;
                        }

                        if (RxBytes > 0)
                            break;

                        // wait for bytes to be received
                        GET_TIME
                        MUTEX_LOCK
                        TIMED_WAIT
                        MUTEX_UNLOCK

                        Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                        debug_print("RxBytes    = %d\n", RxBytes);
                        debug_print("TxBytes    = %d\n", TxBytes);
                        debug_print("EventDWord = %08x\n", EventDWord);

                        if (Status != FT_OK)
                        {
                            debug_print("FT232 get status failed!\n");
                            return Status = DRIVER_ERROR;
                        }
                        else if (RxBytes)
                        {
                            debug_print("FT232 status returned: RxBytes = %d bytes\n", RxBytes);

                            Status = FT_Read(i440Handle, pValBuf + byteCtr, RxBytes, &BytesReceived);

                            if ((Status != FT_OK) || (BytesReceived != RxBytes))
                            {
                                debug_print("FT232 FIFO read of %d bytes failed!\n", RxBytes);
                                Status = DRIVER_ERROR;
                            }
                            else
                                debug_print("Read of %d bytes successful!\n", BytesReceived);

                            byteCtr += (UINT16)BytesReceived;
                        }
                    }
                } while (byteCtr < 24);

                if (Status == SUCCESS)
                {
                    for (unsigned int i = 0; i < 8; i++)
                    {
                        debug_print("FT232 conversion data:\n");
                        debug_print("  Sample %d\n", i + 1);
                        debug_print("  D15 - D0 = 0x%04x\n", (*(pValBuf + (i * 3)) << 8) | *(pValBuf + (i * 3) + 1));
                        debug_print("  Channel  = %d\n", (*(pValBuf + (i * 3) + 2) >> 3) & 0x7);
                        debug_print("  SoftSpan = %d\n", *(pValBuf + (i * 3) + 2) & 0x7);
                    }
                }
            }
        }

        return Status;
    }

    int GetChannelSequence(unsigned char *pValBuf, unsigned int *pChanSeq)
    {
        UINT8 writeBuf[MAX_SEQ_SIZE] = { 0 };
        DWORD byteCtr = 0;
        DWORD cnt = 0;
        DWORD seq = 0;
        DWORD TxBytes = 0;
        DWORD RxBytes = 0;
        DWORD BytesReceived = 0;
        DWORD EventDWord = 0;
        DWORD EventMask = FT_EVENT_RXCHAR;
        timespec tout;
        int Status = SUCCESS;

        if ((pValBuf == NULL) || (pChanSeq == NULL))
        {
            Status = INVALID_PARAMETER;
        }
        else
        {
            // verify pChanSeq values are valid
            for (int i = 0; i <= MAX_SEQ_SIZE; i++)
            {
                if (*(pChanSeq + i) == 0xff)
                    break;

                if (*(pChanSeq + i) >= ADC_NO_CHANNEL || i == MAX_SEQ_SIZE)
                    return Status = INVALID_SEQUENCE;
            }

            // flush FIFO
            Status = FlushRxQueue();

            if (Status != FT_OK)
            {
                debug_print("FT232 FIFO flush failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_mutex_init(&eh.eMutex, NULL);

            if (Status != FT_OK)
            {
                debug_print("Mutex initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_cond_init(&eh.eCondVar, NULL);

            if (Status != FT_OK)
            {
                debug_print("Condition variable initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = FT_SetEventNotification(i440Handle, EventMask, &eh);

            if (Status != FT_OK)
            {
                debug_print("FT232 set event notification failed!\n");
                return Status = DRIVER_ERROR;
            }

            // build control words and start conversions
            do {
                writeBuf[cnt++] = (UINT8)(VALID_BIT | (*(pChanSeq + seq) << 3) | chanSoftSpan[*(pChanSeq + seq)]);
                seq++;
            } while (*(pChanSeq + seq) != 0xff);

            Status = FT_Write(i440Handle, writeBuf, cnt, &BytesReceived);

            if ((Status != FT_OK) || (BytesReceived != cnt))
            {
                debug_print("FT232 FIFO write of %d bytes failed!\n", cnt);
                Status = DRIVER_ERROR;
            }
            else
            {
                debug_print("Write of %d bytes successful!\n", BytesReceived);

                do {
                    Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                    if (Status != FT_OK)
                    {
                        err_print("FT232 get status failed!\n");
                        return Status = DRIVER_ERROR;
                    }

                    while (RxBytes == 0)
                    {
                        Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                        if (Status != FT_OK)
                        {
                            err_print("FT232 get status failed!\n");
                            return Status = DRIVER_ERROR;
                        }

                        if (RxBytes > 0)
                            break;

                        // wait for bytes to be received
                        GET_TIME
                        MUTEX_LOCK
                        TIMED_WAIT
                        MUTEX_UNLOCK

                        Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                        debug_print("RxBytes    = %d\n", RxBytes);
                        debug_print("TxBytes    = %d\n", TxBytes);
                        debug_print("EventDWord = %08x\n", EventDWord);

                        if (Status != FT_OK)
                        {
                            debug_print("FT232 get status failed!\n");
                            return Status = DRIVER_ERROR;
                        }
                        else if (RxBytes)
                        {
                            Status = FT_Read(i440Handle, pValBuf + byteCtr, RxBytes, &BytesReceived);

                            if ((Status != FT_OK) || (BytesReceived != RxBytes))
                            {
                                debug_print("FT232 FIFO read of %d bytes failed!\n", RxBytes);
                                Status = DRIVER_ERROR;
                            }
                            else
                                debug_print("Read of %d bytes successful!\n", BytesReceived);

                            byteCtr += (UINT16)BytesReceived;
                        }
                    }
                } while (byteCtr < seq * 3);

                if (Status == SUCCESS)
                {
                    for (unsigned int i = 0; i < cnt; i++)
                    {
                        debug_print("FT232 conversion data:\n");
                        debug_print("  Sample %d\n", i + 1);
                        debug_print("  D15 - D0 = 0x%04x\n", (*(pValBuf + (i * 3)) << 8) | *(pValBuf + (i * 3) + 1));
                        debug_print("  Channel  = %d\n", (*(pValBuf + (i * 3) + 2) >> 3) & 0x7);
                        debug_print("  SoftSpan = %d\n", *(pValBuf + (i * 3) + 2) & 0x7);
                    }
                }
            }
        }

        return Status;
    }

    int GetChannelSequenceCount(unsigned char *pValBuf, unsigned int *pChanSeq, unsigned int *pCount)
    {
        UINT8 writeBuf[20] = { 0 };
        DWORD byteCtr = 0;
        DWORD cnt = 0;
        DWORD seq = 0;
        DWORD seqCount;
        DWORD TxBytes = 0;
        DWORD RxBytes = 0;
        DWORD BytesReceived = 0;
        DWORD EventDWord = 0;
        DWORD EventMask = FT_EVENT_RXCHAR;
        timespec tout;
        BOOL noCnt = FALSE;
        int Status = SUCCESS;

        if ((pValBuf == NULL) || (pChanSeq == NULL) || (pCount == NULL) || (*pCount == 0))
        {
            Status = INVALID_PARAMETER;
        }
        else
        {
            // verify pChanSeq values are valid
            for (int i = 0; i <= MAX_SEQ_SIZE; i++)
            {
                if (*(pChanSeq + i) == 0xff)
                    break;

                if (*(pChanSeq + i) >= ADC_NO_CHANNEL || i == MAX_SEQ_SIZE)
                    return Status = INVALID_SEQUENCE;
            }

            // flush FIFO
            Status = FlushRxQueue();

            if (Status != FT_OK)
            {
                debug_print("FT232 FIFO flush failed!  %d", Status);
                return Status = DRIVER_ERROR;
            }

            // is cnt = 0?
            //if (*pCount == 0)
            //    noCnt = TRUE;

            seqCount = *pCount;

            Status = pthread_mutex_init(&eh.eMutex, NULL);

            if (Status != FT_OK)
            {
                debug_print("Mutex initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_cond_init(&eh.eCondVar, NULL);

            if (Status != FT_OK)
            {
                debug_print("Condition variable initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = FT_SetEventNotification(i440Handle, EventMask, &eh);

            if (Status != FT_OK)
            {
                debug_print("FT232 set event notification failed!\n");
                return Status = DRIVER_ERROR;
            }

            // build control words and start conversions
            writeBuf[cnt++] = CMD_LOAD_SEQ;

            do {
                writeBuf[cnt++] = (UINT8)(VALID_BIT | (*(pChanSeq + seq) << 3) | chanSoftSpan[*(pChanSeq + seq)]);
                seq++;
            } while (*(pChanSeq + seq) != 0xff);

            writeBuf[cnt++] = CMD_SEQ_CNT;
            writeBuf[cnt++] = (UINT8)(seqCount & 0xff);
            writeBuf[cnt++] = (UINT8)((seqCount >> 8) & 0xff);

            debug_print("Write buffer contents:\n");
            for (unsigned int j = 0; j < cnt; j++)
                debug_print(" %02x -", writeBuf[j]);
            debug_print("\n");

            Status = FT_Write(i440Handle, writeBuf, cnt, &BytesReceived);

            if ((Status != FT_OK) || (BytesReceived != cnt))
            {
                debug_print("FT232 FIFO write of %d bytes failed!\n", cnt);
                Status = DRIVER_ERROR;
            }
            else
            {
                debug_print("Write of %d bytes successful!\n", BytesReceived);

                do {
                    Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);
            
                    if (Status != FT_OK)
                    {
                        err_print("FT232 get status failed!\n");
                        return Status = DRIVER_ERROR;
                    }
                
                    while (RxBytes == 0)
                    {
                        Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                        if (Status != FT_OK)
                        {
                                err_print("FT232 get status failed!\n");
                                return Status = DRIVER_ERROR;
                        }

                        if (RxBytes > 0)
                            break;
 
                        // wait for bytes to be received
                        GET_TIME
                        MUTEX_LOCK
                        TIMED_WAIT
                        MUTEX_UNLOCK
                    }

                    Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                    debug_print("RxBytes    = %d\n", RxBytes);
                    debug_print("TxBytes    = %d\n", TxBytes);
                    debug_print("EventDWord = %08x\n", EventDWord);

                    if (Status != FT_OK)
                    {
                        debug_print("FT232 get status failed!\n");
                        return Status = DRIVER_ERROR;
                    }
                    else if (RxBytes)
                    {
                        Status = FT_Read(i440Handle, pValBuf + byteCtr, RxBytes, &BytesReceived);

                        if ((Status != FT_OK) || (BytesReceived != RxBytes))
                        {
                            debug_print("FT232 FIFO read of %d bytes failed!\n", RxBytes);
                            Status = DRIVER_ERROR;
                        }
                        else
                            debug_print("Read of %d bytes successful!\n", BytesReceived);

                        byteCtr += (UINT16)BytesReceived;
                    }
                } while ((byteCtr < *pCount * 3) || noCnt);

                if (Status == SUCCESS)
                {
                    for (unsigned int i = 0; i < *pCount; i++)
                    {
                        debug_print("FT232 conversion data:\n");
                        debug_print("  Sample %d\n", i + 1);
                        debug_print("  D15 - D0 = 0x%04x\n", (*(pValBuf + (i * 3)) << 8) | *(pValBuf + (i * 3) + 1));
                        debug_print("  Channel  = %d\n", (*(pValBuf + (i * 3) + 2) >> 3) & 0x7);
                        debug_print("  SoftSpan = %d\n", *(pValBuf + (i * 3) + 2) & 0x7);
                    }
                }
            }
        }

        return Status;
    }

    int ConvertToVolts(float *pVoltage, unsigned int channel, int value)
    {
        int Status = SUCCESS;

        if ((pVoltage == NULL) || (channel >= ADC_NO_CHANNEL))
        {
            Status = INVALID_PARAMETER;
        }
        else
        {
            // Convert the raw data to voltage
            value += chanAdjust[channel];
            value &= 0xffff; // remove overflow
            *pVoltage = (value * chanBitVal[channel]);
            *pVoltage += chanOffset[channel];
        }

        return Status;
    }

    int WriteCommands(unsigned char *pValBuf, unsigned char *pCmd, unsigned int cnt)
    {
        UINT8 writeBuf[MAX_CMD_SIZE] = { 0x00 };
        unsigned int i = 0;
        DWORD TxBytes = 0;
        DWORD RxBytes = 0;
        DWORD BytesReceived = 0;
        DWORD totalConvs = 0;
        DWORD EventDWord = 0;
        DWORD EventMask = FT_EVENT_RXCHAR;
        timespec tout;
        int Status = SUCCESS;

        // check parameters
        if (cnt > MAX_CMD_SIZE)
        {
            Status = INVALID_PARAMETER;
        }
        else
        {
            // flush FIFO
            Status = FlushRxQueue();

            if (Status != FT_OK)
            {
                debug_print("FT232 FIFO flush failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_mutex_init(&eh.eMutex, NULL);

            if (Status != FT_OK)
            {
                debug_print("Mutex initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = pthread_cond_init(&eh.eCondVar, NULL);

            if (Status != FT_OK)
            {
                debug_print("Condition variable initialization failed!  %d \n", Status);
                return Status = DRIVER_ERROR;
            }

            Status = FT_SetEventNotification(i440Handle, EventMask, &eh);

            if (Status != FT_OK)
            {
                debug_print("FT232 set event notification failed!\n");
                return Status = DRIVER_ERROR;
            }

            // build command fifo
            for (i = 0; i < cnt; i++)
            {
                writeBuf[i] = *(pCmd + i);

                if (writeBuf[i] >= 0x80) totalConvs++;
            }

            Status = FT_Write(i440Handle, writeBuf, cnt, &BytesReceived);

            if ((Status != FT_OK) || (BytesReceived != (DWORD) cnt))
            {
                debug_print("FT232 FIFO write of %d bytes failed!\n", (UINT16) cnt);
                Status = DRIVER_ERROR;
            }
            else if (pValBuf == NULL)
            {
                debug_print("Write of %d bytes successful!\n", BytesReceived);
            }
            else
            {
                debug_print("Write of %d bytes successful!\n", BytesReceived);

                do {
                    // wait for bytes to be received
                    GET_TIME
                    MUTEX_LOCK
                    TIMED_WAIT
                    MUTEX_UNLOCK

                    Status = FT_GetStatus(i440Handle, &RxBytes, &TxBytes, &EventDWord);

                    if (Status != FT_OK)
                    {
                        debug_print("FT232 get status failed!\n");
                        return Status = DRIVER_ERROR;
                    }
                    else if (RxBytes)
                    {
                        Status = FT_Read(i440Handle, pValBuf, RxBytes, &BytesReceived);

                        if ((Status != FT_OK) || (BytesReceived != RxBytes))
                        {
                            debug_print("FT232 FIFO read of %d bytes failed!\n", RxBytes);
                            Status = DRIVER_ERROR;
                        }
                        else
                            debug_print("Read of %d bytes successful!\n", BytesReceived);
                    }
                } while (!RxBytes);

                Status = StopSequence();
            }
        }

        //CloseHandle(hEvent);

        return Status;
    }

    int StartSequence(void)
    {
        UINT8 writeBuf = 0;
        DWORD BytesReceived = 0;
        int Status = SUCCESS;

        // stop conversions
        writeBuf = CMD_START_AUTO;

        Status = FT_Write(i440Handle, &writeBuf, 1, &BytesReceived);

        if ((Status != FT_OK) || (BytesReceived != 1))
        {
            debug_print("StartSequence command failed!\n");
            Status = DRIVER_ERROR;
        }

        return Status;
    }

    int StopSequence(void)
    {
        UINT8 writeBuf = 0;
        DWORD BytesReceived = 0;
        int Status = SUCCESS;

        // stop conversions
        writeBuf = CMD_STOP_AUTO;

        Status = FT_Write(i440Handle, &writeBuf, 1, &BytesReceived);

        if ((Status != FT_OK) || (BytesReceived != 1))
        {
            debug_print("StopSequence command failed!\n");
            Status = DRIVER_ERROR;
        }

        return Status;
    }

    int FlushRxQueue(void)
    {
        DWORD RxBytes;
        UINT8 readBuf[256] = { 0 };
        DWORD BytesReceived = 0;
        int Status = SUCCESS;

        Status = FT_GetQueueStatus(i440Handle, &RxBytes);

        if (Status != FT_OK)
        {
            FT_Close(i440Handle);
            return Status;
        }
        else if (RxBytes & 0xff)
        {
            Status = FT_Read(i440Handle, readBuf, RxBytes, &BytesReceived);
            debug_print("RX FIFO flushed!\n");
        }

        return Status;
    }
}
